//! Static in-place query construction.
//!
//! Potential interfaces:
//! - `microrm::select().by(KVStore::Key, &key).result(qi)`
//! - `qi.select().by(KVStore::Key, &key).result()`

use crate::{entity::EntityColumn, model::Modelable, Entity, Error};
use std::{marker::PhantomData, collections::HashMap, hash::Hasher};

use super::{WithID, QueryInterface};


#[derive(Debug,PartialEq,Eq,PartialOrd,Ord,Hash)]
pub enum QueryPart {
    Root,
    Where
}

#[derive(Debug)]
pub struct DerivedQuery (HashMap<QueryPart, Vec<String>>);

impl DerivedQuery {
    fn new() -> Self {
        Self { 0: HashMap::new() }
    }

    fn add(mut self, to: QueryPart, what: String) -> Self {
        self.0.entry(to).or_insert_with(|| Vec::new()).push(what);
        self
    }

    fn assemble(mut self) -> String {
        let root = self.0.remove(&QueryPart::Root).unwrap().remove(0);
        let where_ = match self.0.remove(&QueryPart::Where) {
            None => String::new(),
            Some(v) => {
                format!("WHERE {}",
                    v.into_iter().reduce(|a, b| format!("{} AND {}", a, b)).unwrap())
            }
        };

        format!("{} {}",
            root,
            where_
        )
    }
}

trait StaticVersion {
    type Is: 'static;

    fn type_id() -> std::any::TypeId {
        std::any::TypeId::of::<Self::Is>()
    }
}

/// Any query component
trait QueryComponent: StaticVersion {
    fn derive(&self) -> DerivedQuery {
        DerivedQuery::new()
    }

    /// returns the next index to use for binding
    fn bind(&self, stmt: &mut sqlite::Statement<'_>) -> Result<usize, Error>;
}

// helper functions

fn get_stmt<'a, 'l>(qi: &'a QueryInterface<'l>) -> &'a mut sqlite::Statement<'l> {
    todo!()
}

/// Any query that can be completed/executed
trait Resolvable<'r, 'q, T: Entity>: QueryComponent where 'q: 'r {
    fn qi(&self) -> &'r QueryInterface<'q>;

    fn no_result(self) where Self: Sized { todo!() }
    fn result(self) -> Option<WithID<T>> where Self: Sized {
        self.qi().with_cache(
            Self::type_id(),
            || {
                self.qi().db.conn.prepare(self.derive().assemble()).unwrap()
            },
            |stmt| {
                self.bind(stmt);

                self.qi().expect_one_result(stmt, &mut |stmt| {
                    let id: i64 = stmt.read(0).ok()?;
                    Some(WithID::wrap(
                        T::build_from(stmt).ok()?,
                        id,
                    ))
                })
            }
        )
    }
    fn results(self) -> Vec<T> where Self: Sized {
        self.qi().with_cache(
            Self::type_id(),
            || {
                self.qi().db.conn.prepare(self.derive().assemble()).unwrap()
            },
            |stmt| {
                self.bind(stmt);

                todo!()
            }
        )
    }
}

/// Any query that can have a WHERE clause attached to it
trait Filterable<'r, 'q>: Resolvable<'r, 'q, Self::Table>  where 'q: 'r{
    type Table: Entity;

    fn by<C: EntityColumn<Entity = Self::Table>, G: Modelable + ?Sized>(self, col: C, given: &'r G) -> Filter<'r, 'q, Self, C, G> where Self: Sized {
        Filter { wrap: self, col, given, _ghost: PhantomData }
    }
}

pub struct Select<'r, 'q, T: Entity> {
    qi: &'r QueryInterface<'q>,
    _ghost: PhantomData<T>,
}

impl<'r, 'q, T: Entity> Select<'r, 'q, T> {
    pub fn new(qi: &'r QueryInterface<'q>) -> Self {
        Self { qi, _ghost: std::marker::PhantomData }
    }
}

impl <'r, 'q, T: Entity> StaticVersion for Select<'r, 'q, T> {
    type Is = Select<'static, 'static, T>;
}

impl<'r, 'q, T: Entity> QueryComponent for Select<'r, 'q, T> {
    fn derive(&self) -> DerivedQuery {
        DerivedQuery::new().add(QueryPart::Root, format!("SELECT * FROM {}", T::table_name()))
    }

    // next binding point is the first
    fn bind(&self, _stmt: &mut sqlite::Statement<'_>) -> Result<usize, Error> { Ok(1) }
}

impl<'r, 'q, T: Entity> Filterable<'r, 'q> for Select<'r, 'q, T> {
    type Table = T;
}

impl<'r, 'q, T: Entity> Resolvable<'r, 'q, T> for Select<'r, 'q, T> {
    fn qi(&self) -> &'r QueryInterface<'q> { self.qi }
}

/// A concrete WHERE clause
struct Filter<'r, 'q, F: Filterable<'r, 'q>, C: EntityColumn, G: Modelable + ?Sized> where 'q: 'r {
    wrap: F,
    col: C,
    given: &'r G,
    _ghost: PhantomData<&'q ()>
}

impl<'r, 'q, F: Filterable<'r, 'q>, C: EntityColumn, G: Modelable + ?Sized> StaticVersion for Filter<'r, 'q, F, C, G> where <F as StaticVersion>::Is: Filterable<'static, 'static> {
    type Is = Filter<'static, 'static, <F as StaticVersion>::Is, C, u64>;
}

impl<'r, 'q, F: Filterable<'r, 'q>, C: EntityColumn, G: Modelable + ?Sized> QueryComponent for Filter<'r, 'q, F, C, G> where <F as StaticVersion>::Is: Filterable<'static, 'static>, 'q: 'r {
    fn derive(&self) -> DerivedQuery {
        self.wrap.derive().add(QueryPart::Where, format!("{} = ?", self.col.name()))
    }

    fn bind(&self, stmt: &mut sqlite::Statement<'_>) -> Result<usize, crate::Error> {
        let next_index = self.wrap.bind(stmt)?;

        self.given.bind_to(stmt, next_index)?;

        Ok(next_index + 1)
    }
}

impl<'r, 'q, F: Filterable<'r, 'q>, C: EntityColumn, G: Modelable + ?Sized> Resolvable<'r, 'q, C::Entity> for Filter<'r, 'q, F, C, G> where <F as StaticVersion>::Is: Filterable<'static, 'static>, 'q: 'r {
    fn qi(&self) -> &'r QueryInterface<'q> { self.wrap.qi() }
}

impl<'r, 'q, F: Filterable<'r, 'q>, C: EntityColumn, G: Modelable + ?Sized> Filterable<'r, 'q> for Filter<'r, 'q, F, C, G> where <F as StaticVersion>::Is: Filterable<'static, 'static>, 'q: 'r {
    type Table = C::Entity;
}

#[cfg(test)]
mod test_build {
    use microrm_macros::Entity;
    use serde::{Deserialize, Serialize};

    use crate::QueryInterface;

    #[derive(Entity, Serialize, Deserialize)]
    #[microrm_internal]
    pub struct KVStore {
        key: String,
        value: String,
    }

    #[test]
    fn simple_get() {
        use super::*;
        let db = crate::DB::new_in_memory(crate::Schema::new().entity::<KVStore>()).unwrap();
        let qi = db.query_interface();

        assert!(qi.get().by(KVStore::Key, "abc").result().is_none());
    }
}
